@@ -1,14 +1,13 @@
 import copy
-from dataclasses import dataclass
 import json
 from collections import defaultdict
 from pathlib import Path
-import string
-from typing import Set, Dict, Any, List, Type, Optional
-
+from typing import Set, Dict, Any, List, Type, Optional, Tuple
 import tensorflow as tf  # type: ignore
+from keras.engine.keras_tensor import KerasTensor # type: ignore
 from tensorflow.python.keras.utils.generic_utils import func_load  # type: ignore
 
+from keras.engine.node import Node  # type: ignore
 from leap_model_parser.contract.importmodelresponse import DatasetInfo, NodeResponse, WrapperData
 
 layer_attributes_adaptation = {'bias_initializer', 'kernel_initializer'}
@@ -43,75 +42,34 @@ class KerasJsonModelImport:
     def is_representation_block(layer: Dict[str, Any]) -> bool:
         return len(layer['inbound_nodes']) > 1
 
-    def generate_graph(self, model_schema: Dict) -> Dict[str, NodeResponse]:
+    def generate_graph(
+            self, model_schema: Dict, layer_name_to_inbound_nodes: Dict[str, List[Node]]) -> Dict[str, NodeResponse]:
         self.prepare_model_to_import(model_schema)
         for graph_output_layer in model_schema['config']['output_layers']:
             current_layer = self.layer_name_to_layer[graph_output_layer[0]]
-            self.generate_graph_recursive(current_layer, graph_output_layer[1])
+            self.generate_graph_recursive(current_layer, graph_output_layer[1], layer_name_to_inbound_nodes)
 
         nodes = {node.id: node for node in self.nodes_cache.values()}
         _add_origin_name_to_nodes(nodes)
         return nodes
 
-    '''
-    Convert tf ops inbound nodes to the same format as the regular layers, while removing _CONSTANT_VALUE.
-    examples:
-    1.
-        ['conv2d_2', 0, 0, {'slice_spec': [{'start': None, 'stop': None, 'step': None}]}
-        will be converted to:
-        [['conv2d_2', 0, 0]]
-    2.
-        ['tf.__operators__.getitem_1', 0, 0, {'y': ['tf.__operators__.getitem_1', 0, 0]}]
-        will be converted to:
-        [['tf.__operators__.getitem_1', 0, 0], ['tf.__operators__.getitem_1', 0, 0]]
-    3.
-        ['_CONSTANT_VALUE', -1, 1.0, {'y': ['tf.__operators__.add_2', 0, 0], 'name': None}]
-        will be converted to:
-        [['tf.__operators__.add_2', 0, 0]]
-    '''
+    def generate_graph_recursive(self, current_layer: Dict[str, Any], instance_index: int,
+                                 layer_name_to_inbound_nodes: Dict[str, List[Node]]):
+        self.generate_node(current_layer, instance_index, layer_name_to_inbound_nodes)
 
-    @staticmethod
-    def prepare_inbound_nodes(inbound_nodes_to_enumerate: List):
-        inbound_nodes_to_enumerate_prepared = []
-        single_inbound_node_to_enumerate = []
-        for element in inbound_nodes_to_enumerate:
-            if type(element) in (int, str):
-                single_inbound_node_to_enumerate.append(element)
-            elif isinstance(element, dict):
-                if single_inbound_node_to_enumerate and type(single_inbound_node_to_enumerate[0]) == str and \
-                        single_inbound_node_to_enumerate[0] != '_CONSTANT_VALUE':
-                    inbound_nodes_to_enumerate_prepared.append(single_inbound_node_to_enumerate)
-                single_inbound_node_to_enumerate = []
-                for inbound_nodes in element.values():
-                    if isinstance(inbound_nodes, list):
-                        inbound_nodes_to_enumerate_prepared.extend(
-                            KerasJsonModelImport.prepare_inbound_nodes(inbound_nodes))
-        if single_inbound_node_to_enumerate and type(single_inbound_node_to_enumerate[0]) == str:
-            inbound_nodes_to_enumerate_prepared.append(single_inbound_node_to_enumerate)
-
-        return inbound_nodes_to_enumerate_prepared
-
-    def generate_graph_recursive(self, current_layer: Dict[str, Any], instance_index: int):
-        self.generate_node(current_layer, instance_index)
-
-        inbound_nodes = current_layer['inbound_nodes']
+        inbound_nodes = layer_name_to_inbound_nodes[current_layer['config']['name']]
         if len(inbound_nodes) == 0:
             return
-        inbound_nodes_to_enumerate = inbound_nodes[instance_index]
-        if isinstance(inbound_nodes_to_enumerate, dict):
-            inbound_nodes_to_enumerate = list(inbound_nodes_to_enumerate.values())
-        if isinstance(inbound_nodes_to_enumerate[0], list):
-            inbound_nodes_to_enumerate_prepared = []
-            for inbound_node in inbound_nodes_to_enumerate:
-                inbound_nodes_to_enumerate_prepared.extend(self.prepare_inbound_nodes(inbound_node))
-        else:
-            inbound_nodes_to_enumerate_prepared = self.prepare_inbound_nodes(inbound_nodes_to_enumerate)
+        inbound_node = inbound_nodes[instance_index]
+        inbound_layers = [parent_node.layer for parent_node in inbound_node.parent_nodes]
+        for input_index, inbound_layer in enumerate(inbound_layers):
+            input_id = inbound_node.flat_input_ids[input_index]
+            input_node_instance_index = [node.flat_output_ids[0] for node in inbound_layer.inbound_nodes].index(input_id)
+            layer_name = inbound_layer.name
+            input_layer = self.layer_name_to_layer[layer_name]
+            self.generate_node(input_layer, input_node_instance_index, layer_name_to_inbound_nodes)
 
-        for input_index, input_layer_item in enumerate(inbound_nodes_to_enumerate_prepared):
-            input_layer = self.layer_name_to_layer[input_layer_item[0]]
-            self.generate_node(input_layer, input_layer_item[1])
-
-            input_node_name = f"{input_layer['name']}_{input_layer_item[1]}"
+            input_node_name = f"{input_layer['name']}_{input_node_instance_index}"
             current_node_name = f"{current_layer['name']}_{instance_index}"
             connection_name = f"{input_node_name}_{current_node_name}_{input_index}"
             if connection_name in self.visited_connections:
@@ -119,7 +77,8 @@ class KerasJsonModelImport:
 
             self.visited_connections.add(connection_name)
             self.attach(input_layer, input_node_name, current_node_name, input_index)
-            self.generate_graph_recursive(current_layer=input_layer, instance_index=input_layer_item[1])
+            self.generate_graph_recursive(current_layer=input_layer, instance_index=input_node_instance_index,
+                                          layer_name_to_inbound_nodes=layer_name_to_inbound_nodes)
 
     def attach(self, input_layer: Dict, node_input_layer_name: str, node_output_layer_name: str, input_index: int):
         node_output_layer = self.nodes_cache[node_output_layer_name]
@@ -159,9 +118,10 @@ class KerasJsonModelImport:
                 }
         return layer_metadata
 
-    def generate_node(self, layer: Dict[str, Any], instance_index: int):
+    def generate_node(
+            self, layer: Dict[str, Any], instance_index: int, layer_name_to_inbound_nodes: Dict[str, List[Node]]):
         if layer['class_name'] == 'InputLayer':
-            self.handle_input_layer(layer)
+            self._handle_input_layer(layer)
             return
 
         node_key = f"{layer['name']}_{instance_index}"
@@ -175,19 +135,46 @@ class KerasJsonModelImport:
 
         if layer_metadata['type'] == 'wrapper':
             wrapped_layer = self.handle_wrapper_layer(layer)
-            self.generate_node(wrapped_layer, instance_index)
+            self.generate_node(wrapped_layer, instance_index, layer_name_to_inbound_nodes)
             return
 
         if layer['is_representation_block']:
             self.generate_rp_node(layer, layer_metadata, node_key)
         else:
-            self.generate_regular_node(layer, layer_metadata, node_key)
+            self.generate_regular_node(layer, layer_metadata, node_key, layer_name_to_inbound_nodes)
         self.id += 1
 
-    def generate_regular_node(self, layer: Dict[str, Any], layer_metadata: Dict[str, Any], node_key: str):
+    @staticmethod
+    def _serialize_arg(arg):
+        if isinstance(arg, KerasTensor):
+            return arg.name
+        if hasattr(arg, 'numpy'):
+            arg = arg.numpy()
+        if hasattr(arg, 'tolist'):
+            return arg.tolist()
+        if isinstance(arg, (str, float, int, bool, type(None))):
+            return arg
+        if isinstance(arg, list):
+            return [KerasJsonModelImport._serialize_arg(a) for a in arg]
+        if isinstance(arg, dict):
+            return {k: KerasJsonModelImport._serialize_arg(v) for k, v in arg.items()}
+        raise Exception('Unsupported call arg')
+
+    @staticmethod
+    def _serialize_call_args(call_args: Tuple, call_kwargs: Dict) -> str:
+        serialized_call_args = [KerasJsonModelImport._serialize_arg(arg) for arg in call_args]
+        serialized_call_kwargs = {k: KerasJsonModelImport._serialize_arg(v) for k, v in call_kwargs.items()}
+
+        return json.dumps({'serialized_call_args': serialized_call_args,
+                           'serialized_call_kwargs': serialized_call_kwargs})
+
+    def generate_regular_node(self, layer: Dict[str, Any], layer_metadata: Dict[str, Any], node_key: str,
+                              layer_name_to_inbound_nodes: Dict[str, List[Node]]):
         data = layer['config']
         if layer['class_name'] in ('TFOpLambda', 'SlicingOpLambda') or layer['class_name'] in self.custom_layers:
-            data['inbound_nodes'] = layer['inbound_nodes'][0]
+            call_args = layer_name_to_inbound_nodes[layer['config']['name']][0].call_args
+            call_kwargs = layer_name_to_inbound_nodes[layer['config']['name']][0].call_kwargs
+            data['call_args'] = self._serialize_call_args(call_args, call_kwargs)
         elif layer['class_name'] == "Lambda":
             lambda_func = func_load(layer['config']['function'])
             data['arg_names'] = list(lambda_func.__code__.co_varnames)
@@ -307,13 +294,14 @@ class KerasJsonModelImport:
             self.id += 1
         return datasetNode
 
-    def handle_input_layer(self, layer: Dict[str, Any]):
-        datasetNode = self.get_and_add_dataset_node_if_not_existed()
+    def _handle_input_layer(self, layer: Dict[str, Any]):
+        dataset_node = self.get_and_add_dataset_node_if_not_existed()
         if self.dataset_info is None:
-            self.handle_input_layer_without_dataset(layer, datasetNode)
+            self._handle_input_layer_without_dataset(layer, dataset_node)
 
-    def handle_input_layer_without_dataset(self, layer: Dict[str, Any], datasetNode: NodeResponse):
-        inputs = datasetNode.data['datasetVersion']['metadata']['setup']['inputs']
+    @staticmethod
+    def _handle_input_layer_without_dataset(layer: Dict[str, Any], dataset_node: NodeResponse):
+        inputs = dataset_node.data['datasetVersion']['metadata']['setup']['inputs']
 
         if not any(input_data['name'] == layer['name'] for input_data in inputs):
             shape = layer['config']['batch_input_shape']
