@@ -45,7 +45,7 @@ class ModelParser:
 
     def get_keras_model_and_model_graph(
             self, model_path: Path, model_type: ImportModelTypeEnum) -> Tuple[Dict[str, NodeResponse], Optional[Model]]:
-        model_to_keras_converter: Optional[Callable[[str], Tuple[Dict[str, NodeResponse], Optional[Model]]]] = \
+        model_to_keras_converter: Optional[Callable[[str], Tuple[Dict[str, NodeResponse], Model]]] = \
             self._model_types_converter.get(model_type.value)
         if model_to_keras_converter is None:
             raise Exception(f"Unable to import external version, {str(model_path)} file format isn't supported")
@@ -54,7 +54,13 @@ class ModelParser:
         try:
             model_schema, keras_model = model_to_keras_converter(file_path)
             model_generator = KerasJsonModelImport(self.custom_layers, self.dataset_info)
-            graph = model_generator.generate_graph(model_schema)
+
+            layer_name_to_inbound_nodes = {
+                layer.name: layer.inbound_nodes
+                for layer in keras_model.layers
+            }
+
+            graph = model_generator.generate_graph(model_schema, layer_name_to_inbound_nodes)
             return graph, keras_model
         except Exception as e:
             if model_type.value in [ImportModelTypeEnum.H5_TF2.value, ImportModelTypeEnum.PB_TF2.value]:
@@ -63,10 +69,14 @@ class ModelParser:
                 else:
                     keras_model = self._get_k_model_from_pb_path(file_path)
 
+                layer_name_to_inbound_nodes = {
+                    layer.name: layer.inbound_nodes
+                    for layer in keras_model.layers
+                }
                 rebuilt_model = rebuild_model(keras_model)
                 model_schema, keras_model = self.convert_to_keras_model(rebuilt_model)
                 model_generator = KerasJsonModelImport(self.custom_layers, self.dataset_info)
-                graph = model_generator.generate_graph(model_schema)
+                graph = model_generator.generate_graph(model_schema, layer_name_to_inbound_nodes)
                 return graph, keras_model
             else:
                 raise e
@@ -93,7 +103,7 @@ class ModelParser:
     def convert_json_model(cls, file_path: str) -> Tuple[Dict[str, NodeResponse], None]:
         with open(file_path, 'r') as f:
             model_schema = json.load(f)
-        return model_schema, None
+        return model_schema, tf.keras.models.model_from_json(json.dumps(model_schema))
 
     def convert_pb_model(self, file_path: str) -> Tuple[Dict[str, NodeResponse], Model]:
         k_model = self._get_k_model_from_pb_path(file_path)
